import { TransformationPath } from "@upload-io/upload-shared-iso";

export class TransformationUtils {
  static resolveTransformation<T extends TransformationPath>(path: TransformationPath, transformations: T[]): T[] {
    const stripWildScope = (wildScope: string): string => wildScope.slice(0, -1);

    const fullyWildScopesIntersect = (leftWild: string, rightWild: string): boolean => {
      const left = stripWildScope(leftWild);
      const right = stripWildScope(rightWild);
      return left.startsWith(right) || right.startsWith(left);
    };

    const semiWildScopesIntersect = (wildScope: string, absoluteScope: string): boolean =>
      absoluteScope.startsWith(stripWildScope(wildScope));

    const scopesIntersect = (left: string, right: string): boolean => {
      const leftWild = left.endsWith("*");
      const rightWild = right.endsWith("*");
      if (leftWild && rightWild) {
        return fullyWildScopesIntersect(left, right);
      } else if (leftWild) {
        return semiWildScopesIntersect(left, right);
      } else if (rightWild) {
        return semiWildScopesIntersect(right, left);
      }
      return left === right;
    };

    const scopeSetsIntersect = (left: string[], right: string[]): boolean =>
      left.some(l => right.some(r => scopesIntersect(l, r)));

    const slugsIntersect = (left: string, right: string): boolean => left === right; // This might become more complex in future.

    return transformations.filter(
      x =>
        slugsIntersect(x.slug, path.slug) &&
        scopeSetsIntersect(x.scope.tag, path.scope.tag) &&
        scopeSetsIntersect(x.scope.mime, path.scope.mime)
    );
  }
}
